@@ -15,4 +15,8 @@ extern UBYTE PORTB;
 
 extern int xe_bank;
 
+void PIA_Initialise (int *argc, char *argv[]);
+UBYTE PIA_GetByte (UWORD addr);
+int PIA_PutByte (UWORD addr, UBYTE byte);
+
 #endif
